<?php

namespace chillerlan\FastRoute\DataGenerator;

use chillerlan\FastRoute\BadRouteException;

abstract class DataGeneratorAbstract implements DataGeneratorInterface{

	// Semi-accurate detection for capturing groups
	const CGROUP_REGEX =
		'~
			(?:
			    \(\?\(
			  | \[ [^\]\\\\]* (?: \\\\ . [^\]\\\\]* )* \]
			  | \\\\ .
			) (*SKIP)(*FAIL) |
			\(
			(?!
			    \? (?! <(?![!=]) | P< | \' )
			  | \*
			)
		~x';

	protected $approxChunkSize = 0; // friendly reminder
	protected $staticRoutes = [];
	protected $methodToRegexToRoutesMap = [];

	public function addRoute(string $httpMethod, array $routeData, $handler){
		count($routeData) === 1 && is_string($routeData[0])
			? $this->addStaticRoute($httpMethod, $routeData, $handler)
			: $this->addVariableRoute($httpMethod, $routeData, $handler);
	}

	public function getData():array{
		return empty($this->methodToRegexToRoutesMap)
			? [$this->staticRoutes, []]
			: [$this->staticRoutes, $this->generateVariableRouteData()];
	}

	private function generateVariableRouteData():array{
		$data = [];

		foreach($this->methodToRegexToRoutesMap as $method => $regexToRoutesMap){
			$count = count($regexToRoutesMap);
			$data[$method] = array_map([$this, 'processChunk'], array_chunk(
				$regexToRoutesMap,
				ceil($count / max(1, round($count / $this->approxChunkSize))),
				true
			));
		}

		return $data;
	}

	private function addStaticRoute(string $httpMethod, array $routeData, $handler){

		if(isset($this->staticRoutes[$httpMethod][$routeData[0]])){
			throw new BadRouteException(sprintf(
				'Cannot register two routes matching "%s" for method "%s"',
				$routeData[0],
				$httpMethod
			));
		}

		if(isset($this->methodToRegexToRoutesMap[$httpMethod])){
			/** @var Route $route */
			foreach($this->methodToRegexToRoutesMap[$httpMethod] as $route){
				if($route->matches($routeData[0])){
					throw new BadRouteException(sprintf(
						'Static route "%s" is shadowed by previously defined variable route "%s" for method "%s"',
						$routeData[0],
						$route->regex,
						$httpMethod
					));
				}
			}
		}

		$this->staticRoutes[$httpMethod][$routeData[0]] = $handler;
	}

	private function addVariableRoute(string $httpMethod, array $routeData, $handler){
		list($regex, $variables) = $this->buildRegexForRoute($routeData);

		if(isset($this->methodToRegexToRoutesMap[$httpMethod][$regex])){
			throw new BadRouteException(sprintf(
				'Cannot register two routes matching "%s" for method "%s"',
				$regex,
				$httpMethod
			));
		}

		$this->methodToRegexToRoutesMap[$httpMethod][$regex] = new Route($httpMethod, $handler, $regex, $variables);
	}

	private function buildRegexForRoute(array $routeData):array{
		$regex = '';
		$variables = [];

		foreach($routeData as $part){

			if(is_string($part)){
				$regex .= preg_quote($part, '~');
				continue;
			}
			
			if(isset($variables[$part[0]])){
				throw new BadRouteException(sprintf(
					'Cannot use the same placeholder "%s" twice',
					$part[0]
				));
			}

			if($this->regexHasCapturingGroups($part[1])){
				throw new BadRouteException(sprintf(
					'Regex "%s" for parameter "%s" contains a capturing group',
					$part[1],
					$part[0]
				));
			}

			$variables[$part[0]] = $part[0];
			$regex .= '('.$part[1].')';
		}

		return [$regex, $variables];
	}

	private function regexHasCapturingGroups(string $regex):bool{

		if(strpos($regex, '(') === false){
			// Needs to have at least a ( to contain a capturing group
			return false;
		}

		return preg_match(self::CGROUP_REGEX, $regex);
	}

}
